<?php

namespace Shasoft\SamoyedCMG\Generator\Trait;

use Shasoft\Support\FriendProxy;
use Shasoft\SamoyedCMG\Generator\ManagerGenerators;
use Shasoft\SamoyedCMG\Generator\CodeGenerator as CodeGeneratorObject;
use Shasoft\SamoyedCMG\Generator\RouteGenerator;
use Shasoft\SamoyedCMG\Service\IService;

// Методы для Пути и Маршрута
trait TraitCodeGenerator
{
    // Маршрут
    private ?string $route = null;
    // Имя маршрута
    private ?string $name = null;
    // HTTP метод
    private ?string $httpMethod = null;
    // Сервисы
    private array $services = [];
    // Скрипты
    private array $scripts = [];
    // Приоритеты посредников
    private array $middlewarePriorities = [];
    // Список посредников
    private array $middlewares = [];
    // Список создателей
    private array $generators = [];
    // Установить HTTP метод
    public function setHttpMethod(?string $httpMethod): static
    {
        $this->httpMethod = $httpMethod;
        // Вернуть указатель на себя
        return $this;
    }
    // Маршруты
    public function routes(?\Closure $cb = null): array
    {
        $ret  = [];
        foreach ($this->childs as $id) {
            // Получить маршруты
            $routes = ManagerGenerators::get($id)->routes($cb);
            // Объединить с результатом
            $ret = array_merge($ret, $routes);
        }
        if ($this instanceof RouteGenerator) {
            $ret[] = $this;
        }
        return $ret;
    }
    // Получить значение параметра иерархически
    private function getHierProperyValue(string $name, mixed $ret, \Closure $mergeFn): mixed
    {
        $obj = $this;
        while (!is_null($obj)) {
            //s_print('<title>' . $obj->id() . '</> => ' . get_class($obj)); s_dump($obj);
            // Создать прокси
            $objProxy = new FriendProxy($obj);
            // Текущее значение
            $value = $objProxy->{$name};
            //s_dump($value);
            // Объединить значения
            $ret = $mergeFn($value, $ret);
            // Перейти к вышестоящему объекту
            $parentId = $obj->parentId();
            $obj = $parentId != 0 ? ManagerGenerators::get($parentId) : null;
        }
        return $ret;
    }
    // Получить HTTP метод
    public function getHttpMethod(): string
    {
        $values = $this->getHierProperyValue('httpMethod', [], function ($parentValue, $childValue) {
            $childValue[] = $parentValue;
            return $childValue;
        });
        foreach ($values as $value) {
            if (!is_null($value)) {
                return $value;
            }
        }
        return 'GET';
    }
    // Получить сервисы
    public function getServices(): array
    {
        return $this->getHierProperyValue('services', [], function ($parentValues, $childValues) {
            $ret = $childValues;
            foreach ($parentValues as $name => $classname) {
                if (!array_key_exists($name, $ret)) {
                    $ret[$name] = $classname;
                }
            }
            return $ret;
        });
    }
    // Получить скрипты
    public function getScripts(): array
    {
        // Список типов
        $typesAll = [];
        // Список скриптов
        $scriptItems = [];
        foreach (ManagerGenerators::scripts() as $scriptItem) {
            $scriptItems[$scriptItem['id']] = $scriptItem;
        }
        // Получить сервисы
        return array_values($this->getHierProperyValue('scripts', [], function ($parentValues, $childValues) use (&$typesAll, &$scriptItems) {
            $ret = $childValues;
            foreach (array_keys($parentValues) as $id) {
                if (!array_key_exists($id, $ret)) {
                    // Добавлять скрипт?
                    $hasAdd = true;
                    // Информация о скрипте
                    $scriptItem = $scriptItems[$id];
                    foreach ($scriptItem['types'] as $typename => $type) {
                        // Этот тип уже добавлялся?
                        if (array_key_exists($typename, $typesAll)) {
                            $hasAdd = false;
                            break;
                        }
                        $typesAll[$typename] = $scriptItem['filepath'];
                    }
                    // Нужно добавлять?
                    if ($hasAdd) {
                        $ret[$id] = $scriptItem; //['filepath'];
                    }
                }
            }
            return $ret;
        }));
    }
    // Получить приоритеты посредников
    public function getMiddlewarePriorities(): array
    {
        return $this->getHierProperyValue('middlewarePriorities', [], function ($parentValues, $childValues) {
            $ret = $childValues;
            foreach ($parentValues as $name => $priority) {
                if (!array_key_exists($name, $ret)) {
                    $ret[$name] = $priority;
                }
            }
            return $ret;
        });
    }
    // Получить посредников
    public function getMiddlewares(): array
    {
        // Приоритеты
        $prioritys = $this->getMiddlewarePriorities();
        //
        $index = 0;
        // Посредники
        $ret = $this->getHierProperyValue('middlewares', [], function ($parentValues, $childValues) use (&$prioritys, &$index) {
            $ret = $childValues;
            foreach ($parentValues as $item) {
                // Индекс
                $item['index'] = ($index++);
                // Приоритет
                $classname = $item['classname'];
                if (array_key_exists($classname, $prioritys)) {
                    $item['priority'] = $prioritys[$classname];
                }
                // Добавить
                $ret[] = $item;
            }
            return $ret;
        });
        // Сортировать
        usort($ret, function (array $item1, array $item2) {
            if ($item1['priority'] == $item2['priority']) {
                if ($item1['index'] == $item2['index']) {
                    return 0;
                }
                return $item1['index'] < $item2['index'] ? -1 : 1;
            }
            return $item1['priority'] < $item2['priority'] ? -1 : 1;
        });
        return $ret;
    }
    // Установить сервис
    public function setService(string $name, string $classname): bool
    {
        return $this->_setService(new \ReflectionClass($name), $classname);
    }
    // Проверить наличие сервиса
    public function hasService(string $name): bool
    {
        return array_key_exists($name, $this->services);
    }
    // Получить значение сервиса
    public function getService(string $name): string
    {
        return $this->services[$name];
    }
    private function _setService(\ReflectionClass $refClass, string $classname): bool
    {
        $ret = false;
        // Имя сервиса
        $name = $refClass->getName();
        if ($name != IService::class) {
            // Если такой сервис ещё не установлен
            if (!array_key_exists($name, $this->services)) {
                //
                $ret = true;
                // установить его
                $this->services[$name] = $classname;
                // Проверить дочерние интерфейсы
                foreach ($refClass->getInterfaces() as $refInterface) {
                    $this->_setService($refInterface, $classname);
                }
                // Установить флаг изменений
                $this->setModify();
            }
        }
        return $ret;
    }
    // Установить приоритеты посредников
    public function setMiddlewarePriorities(array $priorities): static
    {
        foreach ($priorities as $name => $priority) {
            $this->middlewarePriorities[$name] = $priority;
        }
        // Вернуть указатель на себя
        return $this;
    }
    // Добавить посредник
    public function addMiddleware(string $classname, array $args = [], int $priotity = 0): static
    {
        // Индекс добавляемого посредникаxs`        
        $index = count($this->middlewares);
        // Добавить посредник в список
        $this->middlewares[] = [
            'priority' => $priotity,
            'index' => $index,
            'classname' => $classname,
            'args' => $args
        ];
        // Установить флаг изменений
        $this->setModify();
        // Вернуть указатель на себя
        return $this;
    }
    // Фильтрация посредников
    public function filterMiddleware(?\Closure $cb = null): static
    {
        if (is_null($cb)) {
            if (!empty($this->middlewares)) {
                // Установить флаг изменений
                $this->setModify();
                // Удалить все посредники
                $this->middlewares = [];
            }
        } else {
            // Отфильтровать
            $middlewares = array_filter($this->middlewares, function (array $item) use ($cb) {
                return $cb($item['classname'], $item['args']);
            });
            // Если что-то отфильтровали
            if (count($middlewares) != count($this->middlewares)) {
                // Установить флаг изменений
                $this->setModify();

                // Отсортировать
                usort($middlewares, function (array $itemA, array $itemB) {
                    if ($itemA['index'] == $itemB['index']) {
                        return 0;
                    }
                    return $itemA['index'] < $itemB['index'] ? 1 : -1;
                });
                // Перенумеровать
                $this->middlewares = [];
                $index = 0;
                foreach ($middlewares as $middleware) {
                    $middleware['index'] = $index;
                    $this->middlewares[] = $middleware;
                    $index++;
                }
            }
        }
        // Вернуть указатель на себя
        return $this;
    }
    // Добавить создатель 
    public function addGenerator(CodeGeneratorObject $generator): static
    {
        //s_dump($generator->id() . ' => ' . $this->id());
        $this->generators[] = $generator->id();
        // Вернуть указатель на себя
        return $this;
    }
    // Получить генераторы
    public function getGenerators(): array
    {
        return $this->getHierProperyValue('generators', [], function ($parentValues, $childValues) {
            return array_merge($parentValues, $childValues);
        });
    }
    // Настроить генератор
    public function tune(\Closure $cb): bool
    {
        $ret = false;
        $refFunction = new \ReflectionFunction($cb);
        $args = $refFunction->getParameters();
        if (empty($args)) {
            throw new \Exception('Не указан сервис-аргумент для настройки');
        }
        $arg = $args[0];
        if (!$arg->hasType()) {
            throw new \Exception('Не указан тип сервиса для настройки');
        }
        $classname = (string)$arg->getType();
        $generatorContainer = $this;
        $level = 0;
        $rc = null;
        do {
            $generatorContainerProxy = new FriendProxy($generatorContainer);
            $index = 0;
            foreach ($generatorContainerProxy->generators as $id) {
                $generator = ManagerGenerators::get($id);
                if ($generator instanceof $classname) {
                    // Ok
                    $ret = true;
                    // Вызвать функцию настройки
                    $rc = $cb($generator, $index, $level);
                    // Если в создателе указано что нужно настраивать только один объект
                    if ((new FriendProxy($generator))->hasOneTune()) {
                        // то дальше можно не продолжать
                        $rc = false;
                    }
                    // Если дальше можно не продолжать
                    if ($rc === false) {
                        // то выходим из цикла
                        break;
                    }
                }
            }
            // Перейти к родителю
            $generatorContainer = $generatorContainerProxy->parentId() != 0 ? ManagerGenerators::get($generatorContainerProxy->parentId()) : null;
            // Увеличить уровень
            $level++;
        } while (!is_null($generatorContainer) && $rc !== false);
        // Нашелся хоть один объект?
        return $ret;
    }
    // Запустить генерацию
    private function generate(): void
    {
        // Выполнить все генерации кода
        foreach ($this->generators as $id) {
            ManagerGenerators::runMake($id, $this->id());
        }
        // Выполнить генерацию дочерних объектов
        foreach ($this->childs as $id) {
            (new FriendProxy(ManagerGenerators::get($id)))->generate();
        }
    }
}
