<?php

namespace Shasoft\Support;

use Shasoft\Support\Php;
use Shasoft\Support\File;
use Shasoft\Support\Cache;
use Shasoft\Support\Console;

// Работа с менеджером пакетов COMPOSER
// https://phpprofi.ru/blogs/post/52
// https://getcomposer.ucoz.org/publ/kniga/skhema_composer_json/2-1-0-5
class Composer
{
    // Имя файла с информацией о пакете
    const FILEINFO = 'composer.json';
    // Директория с пакетами
    protected string $basepath;
    // Список локальных пакетов
    protected array $localPackages = [];
    // Объект КЕШа
    protected Cache $cache;
    // Классы из установленных пакетов
    private static bool $has_autoload_register = false;
    private static array $package_map = [];
    private static array $classes_map = [];
    // Конструктор
    public function __construct(string $filepath = null, $extCache = false, protected $autoload = false)
    {
        // Если не указана
        if (is_null($filepath)) {
            // то берем текущую
            $filepath = __DIR__ . '/../../../../';
        }
        // Базовая директория
        $this->basepath = File::normalize($filepath, true);
        // Создать объект КЕШа
        $this->cache = new Cache($extCache == false ? false : $this->path(null, 'cache.' . $extCache));
        // Сгенерировать список локальных пакетов
        $composer = File::load($this->basepath . '/' . self::FILEINFO, []);
        if (array_key_exists('repositories', $composer)) {
            foreach ($composer['repositories'] as $item) {
                if ($item['type'] == 'path') {
                    $pathComposer = File::normalize($this->basepath . '/' . $item['url'], true) . '/' . self::FILEINFO;
                    $info = File::load($pathComposer, false);
                    if ($info !== false) {
                        $this->localPackages[$info['name']] = dirname($pathComposer);
                    }
                }
            }
        }
        // Установить загрузчик установленных пакетов (если он ещё не установлен)
        self::set_autoload_register();
    }
    // Установить загрузчик установленных пакетов (если он ещё не установлен)
    static private function set_autoload_register(): void
    {
        // Установить загрузчик установленных пакетов (если он ещё не установлен)
        if (self::$has_autoload_register === false) {
            // Загрузчик установлен
            self::$has_autoload_register = true;
            // Зарегистрировать новый обработчик
            spl_autoload_register(function ($classname) {
                // Класс есть в списке?
                if (array_key_exists($classname, self::$classes_map)) {
                    // Подключить файл с классом
                    require_once self::$classes_map[$classname];
                }
            });
        }
    }
    // Подключить файлы автозагрузки
    public function autoload(): static
    {
        if (!$this->autoload) {

            // Подключить файл автозагрузки
            require $this->path(null, 'autoload.php');
            // Автозагрузка подключалась
            $this->autoload = true;
        }
        // Вернуть указатель на себя
        return $this;
    }
    // Список локальных пакетов
    public function localPackages()
    {
        return $this->localPackages;
    }
    // Получить директорию пакета
    public function path($packageName = null, $pathX = null)
    {
        return File::normalize($this->basepath . '/vendor' . (is_null($packageName) ? '' : '/' . $packageName) . (empty($pathX) ? '' : '/' . $pathX), true);
    }
    // Получить информацию о пакета
    public function info($packageName)
    {
        return $this->cache->get('info', $packageName, function ($packageName) {
            return File::load($this->path($packageName, self::FILEINFO), false);
        });
    }
    // Это локальный пакет?
    public function hasLocal($packageName): bool
    {
        return array_key_exists($packageName, $this->localPackages);
    }
    // Список пакетов
    public function packages()
    {
        return $this->cache->get('packages', '', function () {
            $ret = [];
            // Список компаний
            $vendors = glob($this->path('/*'), GLOB_ONLYDIR);
            foreach ($vendors as $vendorPath) {
                // Получить список пакетов
                $packages = glob($vendorPath . '/*', GLOB_ONLYDIR);
                // Перебрать все пакеты
                foreach ($packages as $packagePath) {
                    $composer = File::load($packagePath . '/' . self::FILEINFO, false);
                    if ($composer !== false) {
                        $ret[$composer['name']] = File::normalize($packagePath, true);
                    }
                }
            }
            return $ret;
        });
    }
    // Получить дочерние пакета
    public function childs($packageName)
    {
        return $this->cache->get('childs', $packageName, function ($packageName) {
            $ret = [];
            $composer = $this->info($packageName);
            if ($composer !== false) {
                // Добавить все дочерние пакеты
                if (array_key_exists('require', $composer)) {
                    foreach ($composer['require'] as $name => $version) {
                        if (strpos($name, '/') !== false) {
                            $ret[] = $name;
                        }
                    }
                }
            }
            return $ret;
        });
    }
    // Получить ВСЕ дочерние пакета (т.е. и дочерние дочерних)
    protected function _childsAll(&$packages, $packageNames)
    {
        foreach ($packageNames as $packageName) {
            // Если данный пакет ещё не обрабатывался
            if (!array_key_exists($packageName, $packages)) {
                // Получить дочерние пакеты
                $childPackages = $this->childs($packageName);
                // ДобавитьsubPackage
                $packages[$packageName] = [
                    'name' => $packageName,
                    'parent' => [],
                    'priority' => 0,
                    'require' => $childPackages
                ];
                // Обработать дочерние пакеты
                $this->_childsAll($packages, $childPackages);
            }
        }
    }
    // Обновить приоритет
    protected function _updatePriority(&$packages, $packageName, $subPackage, &$ok)
    {
        // Обновлять только если по данному пакету ещё не было обновлений
        if (!array_key_exists($packageName, $ok)) {
            // Выставить флаг что было обновление приоритета
            $ok[$packageName] = 1;
            // Увеличить приоритет
            if (!empty($subPackage)) {
                $packages[$packageName]['parent'][] = $subPackage;
                $packages[$packageName]['priority']++;
            }
            // Увеличить приоритет для всех пакетов, от которых этот пакет зависит
            foreach ($packages[$packageName]['require'] as $name) {
                $this->_updatePriority($packages, $name, $packageName, $ok);
            }
        }
    }
    // Получить ВСЕ дочерние пакета (т.е. и дочерние дочерних)
    public function childsAll($packageNames)
    {
        // Отсортировать
        sort($packageNames, SORT_STRING);
        // Получить список дочерних пакетов
        return $this->cache->get('childsAll', implode(',', $packageNames), function ($s)  use ($packageNames) {
            $packages = [];
            // Обработать все пакеты
            $this->_childsAll($packages, $packageNames);
            // Определить приоритеты
            foreach ($packages as $packageName => $packageInfoi) {
                $ok = [];
                $this->_updatePriority($packages, $packageName, null, $ok);
            }
            // Получить массив инфомации о пакетах
            $packages = array_values($packages);
            // Отсортировать
            usort($packages, function ($a, $b) {
                if ($a['priority'] < $b['priority']) return 1; // если первое значение больше второго
                if ($a['priority'] > $b['priority']) return -1; // если первое значение меньше второго        

                return 0;
            });
            // Сформировать список имен пакетов
            $ret = [];
            foreach ($packages as $package) {
                $ret[] = $package['name'];
            }
            return $ret;
        });
    }
    // Перебрать все элементы(файлы/директории) пакета
    public function items(string $packageName, \Closure $cb = null)
    {
        $ret = $this->cache->get('items', $packageName, function ($packageName) use ($cb) {
            $ret = [];
            // Директория пакета
            $pathPackage = $this->path($packageName);
            // Перебрать все файла директории
            File::items($pathPackage, function (\SplFileInfo $spi) use ($cb, &$ret) {
                $ret[] = $spi->getPathname();
            });
            return $ret;
        }, function ($items) {
            return array_map(function ($filepath) {
                return new \SplFileInfo($filepath);
            }, $items);
        }, function ($packageName) {
            // Сохраняем только для НЕлокальных пакетов
            return !$this->hasLocal($packageName);
        });
        // Указана функция обратного вызова?
        if (!is_null($cb)) {
            foreach ($ret as $spi) {
                $cb($spi);
            }
        }
        return $ret;
    }
    // Перебрать все классы пакета
    public function classes(string $packageName, \Closure $cb = null)
    {
        $ret = $this->cache->get('classes', $packageName, function ($packageName) {
            // Получить список элементов
            $items = $this->items($packageName);
            // Определить список PHP Файлов
            $files = [];
            foreach ($items as $spi) {
                // Это файл PHP скрипта?
                if (Php::isPhpFile($spi)) {
                    $files[] = $spi->getPathname();
                }
            }
            //s_print($packageName);
            return array_keys($this->getClassesInFiles($files));
        }, function ($classes) {
            return array_map(function ($classname) {
                return new \ReflectionClass($classname);
            }, $classes);
        }, function ($packageName) {
            // Сохраняем только для НЕлокальных пакетов
            //s_dump($packageName, $this->hasLocal($packageName));
            return !$this->hasLocal($packageName);
        });
        // Указана функция обратного вызова?
        if (!is_null($cb)) {
            foreach ($ret as $refClass) {
                $cb($refClass);
            }
        }
        return $ret;
    }
    // Получить все классы с использованием паралельного процесса
    public function getClassesInFiles(array $files): array
    {
        // Сохранить список файлов
        $filepathFiles = File::tempfile('ser');
        File::save($filepathFiles, $files);
        // Файл с результатом
        $filepathClasses = File::tempfile('ser');

        // Запустить сканирование
        $classes = [];
        do {
            // Запустить поиск
            $rc = Console::exec(Console::phpPath() . ' "' . __DIR__ . '/ComposerGetClassesInFiles.php" "' . $this->path(null, 'autoload.php') . '" "' . $filepathFiles . '" "' . $filepathClasses . '"', null, true, false);
            // Добавить найденные классы
            $classes = array_merge($classes, File::load($filepathClasses, []));
            // Продолжить поиск пока он не завершится успешно
        } while ($rc != 0);
        // Вернуть результат
        return $classes;
    }
    // Получить все классы с использованием паралельного процесса
    public function getClassesAll(): array
    {
        // Список PHP файлов
        $files = [];
        foreach ($this->packages() as $packageName => $path) {
            // Если эт оне локальный пакет
            if (!$this->hasLocal($packageName)) {
                // то проверим наличие в буфере

            }
            foreach ($this->items($packageName) as $spi) {
                // Это файл PHP скрипта?
                if (Php::isPhpFile($spi)) {
                    $files[] = $spi->getPathname();
                }
            }
        }
        // Сохранить список файлов
        $filepathFiles = File::tempfile('ser');
        File::save($filepathFiles, $files);
        // Файл с результатом
        $filepathClasses = File::tempfile('ser');

        // Запустить сканирование
        $classes = [];
        do {
            // Запустить поиск
            $rc = Console::exec(Console::phpPath() . ' "' . __DIR__ . '/ComposerGetClassesInFiles.php" "' . $this->path(null, 'autoload.php') . '" "' . $filepathFiles . '" "' . $filepathClasses . '"', null, true, false);
            // Добавить найденные классы
            $classes = array_merge($classes, File::load($filepathClasses, []));
            // Продолжить поиск пока он не завершится успешно
        } while ($rc != 0);
        // Вернуть результат
        return $classes;
    }
    // Установить пакеты
    public function install($isFast = false): static
    {
        if ($isFast == false || !file_exists($this->basepath . '/vendor')) {
            Console::exec('composer install', $this->basepath);
        }
        //s_dd($isFast, $this->basepath);
        // Вернуть указатель на себя
        return $this;
    }
    // Установить пакеты
    public function require($packageNames): static
    {
        // Если это просто имя пакета
        if (is_string($packageNames)) {
            // то преобразовать его в массив
            $packageNames = [$packageNames];
        }
        // Установленные в этой сессии пакеты
        $namesRequired = [];
        // 
        $namesRequire = [];
        $namesNoRequire = [];
        foreach ($packageNames as $packageName) {
            // Если папки пакета не существует
            $filepath = $this->path($packageName);
            if (file_exists($filepath)) {
                // то не не устанавливать его
                $namesNoRequire[] = $packageName;
            } else {
                // иначе устанавливать
                $namesRequire[] = $packageName;
                // Добавить пакет в список установленных
                $namesRequired[$packageName] = 1;
            }
        }
        // Если список пакетов для установки не пустой
        if (!empty($namesRequire)) {
            s_dump($namesRequire);
            Console::exec('composer require ' . implode(' ', $namesRequire), $this->basepath);
        }
        //
        foreach ($namesNoRequire as $name) {
            // Пакет уже установлен
            // Но может в пакете есть пакеты, которые туда добавили, но они не установлены?
            //  (такое может быть только при разработке!)
            $childPackages = $this->childs($name);
            foreach ($childPackages as $packageNameChild) {
                // Если пакет не установлен
                if (!file_exists($this->path($packageNameChild))) {
                    // то установить его
                    $this->require($packageNameChild);
                    // Добавить пакет в список установленных
                    $namesRequired[$packageNameChild] = 1;
                }
            }
        }
        // Все классы из вновь установленных пакетов добавить в автозагрузку
        //s_dump($namesRequired);
        foreach ($this->childsAll(array_keys($namesRequired)) as $packageName) {
            //s_dump($packageName);
            // Пакет НЕ проверялся?
            if (!array_key_exists($packageName, self::$package_map)) {
                // Пометить пакет как проверяемы
                self::$package_map[$packageName] = 1;
                // Добавить классы пакета
                $this->items($packageName, function (\SplFileInfo $spi) {
                    if (Php::isPhpFile($spi)) {
                        // Получить список классов файла
                        $classes = Php::getFileClasses($spi->getPathname(), Php::C_ALL, true);
                        // Добавить классы в коллекцию классов
                        foreach ($classes as $classname => $fnExists) {
                            self::$classes_map[$classname] = $spi->getPathname();
                        }
                    }
                });
            }
        }
        // Вернуть указатель на себя
        return $this;
    }
    // Удалить пакеты
    public function remove($packageNames): static
    {
        // Если это просто имя пакета
        if (is_string($packageNames)) {
            // то преобразовать его в массив
            $packageNames = [$packageNames];
        }
        // 
        $names = [];
        foreach ($packageNames as $packageName) {
            // Если пакет не установлен
            $filepath = $this->path($packageName);
            if (file_exists($filepath)) {
                // то его нужно устанавливать
                $names[] = $packageName;
            }
        }
        // Если список пакетов для установки не пустой
        if (!empty($namesRequire)) {
            Console::exec('composer remove ' . implode(' ', $names), $this->basepath);
        }
        // Вернуть указатель на себя
        return $this;
    }
    // Найти имя пакета для сайта и путь файла относительно корня пакета
    public static function findPackage(string $filepath): array|false
    {
        $ret = false;
        // Ищем файл пакета начиная в корне
        $_filepath = File::normalize($filepath);
        $path = dirname($_filepath);
        $lenPrev = 0;
        while (strlen($path) != $lenPrev) {
            // Имя файла с информацией о пакете
            $fileComposer = $path . '/' . self::FILEINFO;
            if (file_exists($fileComposer)) {
                // Читать пакет
                $composer = File::load($fileComposer);
                $ret = [
                    'name' => $composer['name'],
                    'path' => str_replace('\\', '/', $path),
                    'file' => str_replace('\\', '/', substr($_filepath, strlen($path) + 1))
                ];
                // Выйти из цикла
                break;
            }
            // Определим текущую длинну
            $lenPrev = strlen($path);
            // Определим новый путь
            $path = dirname($path);
        }
        //
        return $ret;
    }
}
