<?php

namespace Shasoft\Support;

use Shasoft\Support\Serialize;

class File
{
    // Пропустить ссылки
    public const SKIP_LINKS = 32768;
    // Перебрать все элементы директории
    public static function items(string $filepath, \Closure $cb): void
    {
        // Получить список элементов
        if (file_exists($filepath)) {
            if (is_dir($filepath)) {
                $items = scandir($filepath);
                foreach ($items as $name) {
                    if ($name != '.' && $name != '..') {
                        // Полное имя элемента
                        $filepathItem = $filepath . '/' . $name;
                        // Создать FileInfo
                        $info = new \SplFileInfo($filepathItem);
                        // Вызвать функцию обратного вызова
                        if ($cb($info) !== false) {
                            // Если это директория
                            if ($info->isDir()) {
                                // то вызвать перебор элементов в ней
                                self::items($filepathItem, $cb);
                            }
                        }
                    }
                }
            } else {
                // Вызвать функцию обратного вызова
                $cb(new \SplFileInfo($filepath));
            }
        }
    }
    // Сохранить файл, создав директорию, если она не существует
    public static function save(string $filepath, mixed $data, ?string $format = null, ?int $options = null): int|false
    {
        // Создать директорию
        self::mkdir(dirname($filepath));
        // Определить формат данных
        if (is_null($format)) {
            $format = strtolower(pathinfo($filepath, PATHINFO_EXTENSION));
        }
        // Серилизовать данные в строку
        $str = Serialize::toStr($data, $format);
        // Сохранить данные в файл
        return file_put_contents($filepath, $str);
    }
    // Читать файл
    public static function load(string $filepath, mixed $defaultData = null, ?string $format = null): mixed
    {
        if (file_exists($filepath)) {
            // Читать данные из файла
            $ret = file_get_contents($filepath);
            // Определить формат данных
            if (is_null($format)) {
                $format = strtolower(pathinfo($filepath, PATHINFO_EXTENSION));
            }
            try {
                // Серилизовать данные в строку
                $ret = Serialize::fromStr($ret, $format);
            } catch (\Exception $e) {
                throw new \Exception($e->getMessage() . ' in file ' . $filepath);
            }
        } else {
            if (is_callable($defaultData)) {
                $ret = \call_user_func($defaultData);
                //if( !config('app.debug') ) {
                self::save($filepath, $ret, $format);
                //}
            } else {
                $ret = $defaultData;
            }
        }
        return $ret;
    }
    // Создать имя временного файла
    static $s_tempid = 0;
    public static function tempfile(string $ext = '')
    {
        self::$s_tempid++;
        $id = md5(microtime()) . self::$s_tempid;
        if (!empty($ext)) {
            $ext = '.' . $ext;
        }
        return sys_get_temp_dir() . '/ShasoftFile/' . self::$s_tempid . $id . $ext;
    }
    // Получить относительный путь от файла $from до файла $to
    public static function relative(string $from, string  $to)
    {
        $from = str_replace("\\", '/', $from);
        $to   = str_replace("\\", '/', $to);
        $from = explode('/', $from);
        $to   = explode('/', $to);
        while (!empty($from) && !empty($to) && $from[0] == $to[0]) {
            array_shift($from);
            array_shift($to);
        }
        $ret = './' . str_repeat("../", count($from) - 1) . implode('/', $to);
        return str_replace('\\', '/', $ret);
    }
    // Создать ссылку
    public static function mklink(string $from, string $to): bool
    {
        // Создать директорию (если её нет)
        self::mkdir(dirname($to));
        // Если ссылка существует
        if (file_exists($to) || is_link($to)) {
            // удалить её
            if (is_dir($to)) {
                $rc = @rmdir($to);
                if (!$rc) {
                    if (is_link($to)) {
                        $rc = @unlink($to);
                    }
                }
            } else {
                $rc = @unlink($to);
                if (!$rc) {
                    if (is_link($to)) {
                        $rc = @rmdir($to);
                    }
                }
            }
        }
        // Создать ссылку
        if (!file_exists($from)) {
            s_dd($from, file_exists($from), $to);
        }
        $ret = symlink($from, $to);
        if (!$ret) {
            s_dd($from, $to);
        }
        return $ret;
    }
    // Заменить расширение
    public static function replace_ext(string $filepath, string $ext): string
    {
        $info = pathinfo($filepath);
        return (empty($info['dirname']) ? $info['filename'] : $info['dirname'] . '/' . $info['filename']) . (empty($ext) ? '' : ('.' . $ext));
    }
    // Нормализовать путь. Заменив '\' на '/' и убрав '/' в начале
    public static function posix($path, $removeSlashes = false)
    {
        $ret = str_replace('\\', '/', $path);
        if ($removeSlashes) {
            if (substr($ret, 0, 1) == '/') {
                $ret = substr($ret, 1);
            }
        }
        return $ret;
    }
    // Нормализовать путь. Убрать '..' и '.'
    public static function normalize(string $path, bool $as_posix = false): string
    {
        $path  = str_replace(array('/', '\\'), DIRECTORY_SEPARATOR, $path);
        $parts = array_filter(explode(DIRECTORY_SEPARATOR, $path), 'strlen');
        $ret   = array();
        foreach ($parts as $part) {
            if ('.' == $part) {
                continue;
            }

            if ('..' == $part) {
                array_pop($ret);
            } else {
                $ret[] = $part;
            }
        }
        $ret = implode(DIRECTORY_SEPARATOR, $ret);
        // Нужно нормализовать в формат POSIX
        if ($as_posix) {
            // Преобразовать СЛЭШИ
            $ret = str_replace(DIRECTORY_SEPARATOR, '/', $ret);
            // Может это запуск в Windows?
            if (PHP_OS_FAMILY == 'Windows') {
                if (substr($ret, 1, 1) == ':') {
                    // В начале указанн диск. перевести его в верхний регистр
                    $ret = ucfirst($ret);
                }
            }
        }
        return $ret;
    }
    // Удалить директорию (рекурсивно)
    public static function rmdir(string $path): bool
    {
        //s_dd($path, file_exists($path));
        $ret = true;
        // Директория существует?
        if (file_exists($path)) {
            if (is_dir($path)) {
                // Список элементов
                $items = new \FilesystemIterator($path);
                // Удалить все элементы директории
                foreach ($items as $item) {
                    if (is_dir($item)) {
                        if (is_link($item)) {
                            $ret &= rmdir($item);
                        } else {
                            $ret &= self::rmdir($item);
                        }
                    } else {
                        if (!@unlink($item)) {
                            if (is_link($item)) {
                                $ret &= \rmdir($item);
                            }
                        }
                    }
                }
                // Удалить
                $ret &= rmdir($path);
            } else {
                $ret &= unlink($path);
            }
        }
        return $ret;
    }
    // Вставить в имя файла суффикс
    public static function insSuffiks($filepath, $suffiks)
    {
        // Расширение
        $ext = pathinfo($filepath, PATHINFO_EXTENSION);
        if (!empty($ext)) {
            $ext = '.' . $ext;
        }
        // Имя файла
        $filename = basename($filepath, $ext);
        // Путь
        $path = dirname($filepath);
        if (!empty($path)) {
            $path .= '/';
        }
        // Результат
        return $path . $filename . $suffiks . $ext;
    }
    // Установить функцию трансформации имени файла
    protected static $fnTransform = null;
    public static function setTransformName(\Closure $fn): void
    {
        self::$fnTransform = $fn;
    }
    // Получить трансформированное имя файла
    public static function getTransformName(string $filepath): string
    {
        return is_null(self::$fnTransform) ? $filepath : call_user_func(self::$fnTransform, $filepath);
    }
    // Создать директорию
    public static function mkdir(string $filepath, int $permissions = 0777): bool
    {
        // Проверим существование директорииs
        $ret = file_exists($filepath);
        // Если директория не существует
        if (!$ret) {
            // то создать её
            $ret = mkdir($filepath, $permissions, true);
        }
        return $ret;
    }
}
