<?php

namespace Shasoft\Support;

use PhpParser\Node;
use PhpParser\NodeTraverser;
use PhpParser\NodeVisitorAbstract;
use PhpParser\Node\Stmt\Class_;
use PhpParser\Node\Stmt\Interface_;
use PhpParser\Node\Stmt\Namespace_;
use PhpParser\Node\Stmt\Trait_;
use PhpParser\Node\Stmt\Enum_;
use PhpParser\Node\Stmt\Function_;
use PhpParser\ParserFactory;
use Shasoft\Support\File;

class Php
{
    // Константы
    const C_CLASS     = 1;
    const C_INTERFACE = 2;
    const C_TRAIT     = 4;
    const C_ENUM      = 8;
    const C_FUNCTION  = 16;
    const C_ALL       = self::C_CLASS | self::C_INTERFACE | self::C_TRAIT | self::C_ENUM;
    const C_ALL_FN    = self::C_ALL | self::C_FUNCTION;
    // Получить список классов
    // Для работы необходим пакет nikic/php-parser https://github.com/nikic/PHP-Parser
    protected static $parser = false;
    // Получить список классов
    public static function getClasses($content, $options = self::C_ALL, $fExistsCheck = false)
    {
        $ret = [];
        //s_dump($content);
        // Если парсер не создан
        if (self::$parser === false) {
            // то создать его
            self::$parser = (new ParserFactory)->create(ParserFactory::PREFER_PHP7);
        }
        //s_dd(self::$parser);
        // Получить AST дерево
        try {
            $ast = self::$parser->parse($content);
        } catch (\Exception $e) {
            return false;
        }
        //s_dd($ast);
        // Создать объект для обхода
        $visitor = new class($options) extends NodeVisitorAbstract
        {
            public $namespace = '';
            public $items     = [];
            protected $options;
            // Конструктор
            public function __construct($options)
            {
                // Отключить типы, которые не поддерживаются в данной версии PHP
                if (!function_exists('enum_exists')) {
                    $options &= ~Php::C_ENUM;
                }
                //
                $this->options = $options;
            }
            //
            public function enterNode(Node $node)
            {
                if ($node instanceof Namespace_) {
                    // Clean out the function body
                    if (!is_null($node->name)) {
                        $this->namespace = $node->name->toString();
                    }
                } else {
                    $fn_exists = false;
                    if (($this->options & Php::C_CLASS) && $node instanceof Class_) {
                        $fn_exists = 'class_exists';
                    } else if (($this->options & Php::C_INTERFACE) && $node instanceof Interface_) {
                        $fn_exists = 'interface_exists';
                    } else if (($this->options & Php::C_TRAIT) && $node instanceof Trait_) {
                        $fn_exists = 'trait_exists';
                    } else if (($this->options & Php::C_ENUM) && $node instanceof Enum_) {
                        $fn_exists = 'enum_exists';
                    } else if (($this->options & Php::C_FUNCTION) && $node instanceof Function_) {
                        $fn_exists = 'function_exists';
                    }
                    // Это нужный нам узел?
                    if ($fn_exists !== false) {
                        // И имя определено?
                        if (!is_null($node->name)) {
                            $this->items[] = [
                                'name' => (string) $node->name, //->toString();
                                'fn_exists' => $fn_exists,
                            ];
                        }
                    }
                }
            }
        };
        //
        $traverser = new NodeTraverser();
        $traverser->addVisitor($visitor);
        $ast = $traverser->traverse($ast);
        //
        if (!empty($visitor->namespace)) {
            $visitor->namespace .= '\\';
        }
        // Перебрать все найденные классы
        foreach ($visitor->items as $item) {
            // Полное имя
            $classnameFull = $visitor->namespace . $item['name'];
            //
            if ($item['fn_exists'] == 'function_exists') {
                try {
                    //s_dd($item['fn_exists']($classnameFull), $classnameFull);
                    // Проверить существование (с использованием функции для текущего типа)
                    if ($item['fn_exists']($classnameFull) || $fExistsCheck) {
                        // Записать значение
                        $ret[$classnameFull] = $item['fn_exists'];
                    }
                } catch (\Exception $e) {
                    //var_dump('Exception', $e->getMessage());
                } catch (\Error $e) {
                    //var_dump('Error', $e->getMessage());
                }
            } else {
                //s_dump($classnameFull, $item['fn_exists']);
                //s_dump($item['fn_exists']($classnameFull, true));

                try {
                    // Проверить существование (с использованием функции для текущего типа)
                    if ($fExistsCheck || $item['fn_exists']($classnameFull, true)) {
                        // Записать значение
                        $ret[$classnameFull] = $item['fn_exists'];
                    }
                } catch (\Exception $e) {
                    //var_dump('Exception', $e->getMessage());
                } catch (\Error $e) {
                    //var_dump('Error', $e->getMessage());
                } catch (\Throwable $e) {
                    var_dump('Throwable', $e->getMessage());
                }
            }
        }
        // Вернуть список классов
        return $ret;
    }
    // Получить список классов файла
    public static function getFileClasses($filepath, $options = self::C_ALL, $fExistsCheck = false)
    {
        $ret = [];
        // Файл существует?
        if (file_exists($filepath)) {
            // Читать список классов
            $ret = self::getClasses(file_get_contents($filepath), $options, $fExistsCheck);
        }
        return $ret;
    }
    // Получить тип данных
    // Это стандартный тип?
    static private array $standartTypes = [
        'array' => 1,
        'callable' => 1,
        'bool' => 1,
        'int' => 1,
        'float' => 1,
        'string' => 1,
        'iterable' => 1,
        'void' => 1,
        'object' => 1,
        'null' => 1,
        'false' => 1,
        'mixed' => 1,
        'never' => 1,
        'true' => 1,
    ];
    public static function getType(\ReflectionNamedType|\ReflectionUnionType|\ReflectionIntersectionType|null $type): array
    {
        if (is_null($type)) {
            $ret = [];
        } else {
            $ret = array_map(function ($type) {
                if (substr($type, 0, 1) != '?' && !array_key_exists($type, self::$standartTypes)) {
                    // это класс
                    $type = '\\' . $type;
                }
                return $type;
            }, explode('|', (string)$type));
        }
        return $ret;
    }
    // Получить код прототипа метода
    public static function getCode(\ReflectionMethod $refMethod): string
    {
        $ret = '';
        if ($refMethod->isStatic()) $ret .= ' static';
        if ($refMethod->isPrivate()) $ret .= ' private';
        if ($refMethod->isProtected()) $ret .= ' protected';
        if ($refMethod->isPublic()) $ret .= ' public';
        // Аргументы
        $args = [];
        foreach ($refMethod->getParameters() as $refArg) {
            $arg_s = '';
            $types = self::getType($refArg->getType());
            if (!empty($types)) {
                $arg_s = implode('|', $types) . ' ';
            }
            $arg_s .= '$' . $refArg->getName();
            if ($refArg->isDefaultValueAvailable()) {
                if ($refArg->isDefaultValueConstant()) {
                    $arg_s .= ' = ' . $refArg->getDefaultValueConstantName();
                } else {
                    $arg_s .= ' = ' . var_export($refArg->getDefaultValue(), true);
                }
            }
            $args[] = $arg_s;
        }
        // Прототип
        $ret .= ' function ' . $refMethod->getName() . '(' . implode(', ', $args) . ')';
        // Возвращаемый тип
        if ($refMethod->hasReturnType()) {
            $ret .= ' : ' . implode('|', Php::getType($refMethod->getReturnType()));
        }
        return trim($ret);
    }
    // Получить тип параметра функции
    public static function getTypeFunction(\Closure $cb): array
    {
        $ret = [];
        // Определим тип колонки
        $refFn = new \ReflectionFunction($cb);
        $args = $refFn->getParameters();
        if (!empty($args)) {
            // Типы
            $ret = self::getType($args[0]->getType());
        }
        // Вернуть тип
        return $ret;
    }
    // Проверить наличие объекта
    public static function itemExists(string $name, string $fn_exists, bool $autoload = true): bool
    {
        if ($fn_exists == 'function_exists') {
            return function_exists($name);
        }
        return $fn_exists($name, $autoload);
    }

    /*
    // Информация о типе
    protected static function _getTypeInfo($reflectorType)
    {
        if (is_null($reflectorType)) {
            $ret = false;
        } else {
            $ret = [
                'allowsNull' => $reflectorType->allowsNull(),
                'name'       => $reflectorType->__toString(),
            ];
        }
        return $ret;
    }
    // Информация об атибуте
    protected static function _getAttributesInfo($attributes)
    {
        $ret = [];
        foreach ($attributes as $refAttr) {
            $ret[] = [
                'name' => $refAttr->getName(),
                'args' => $refAttr->getArguments(),
                'isRepeated' => $refAttr->isRepeated(),
            ];
        }
        return $ret;
    }
    // Получить информацию о классе
    public static function getClassInfo($classname)
    {
        //echo $classname . "\n";
        // Создать рефлектор класса
        $reflector = new \ReflectionClass($classname);
        //
        $ret = [
            'classname'      => $reflector->name,
            'filename'       => File::normalize($reflector->getFileName(), true),
            'docComment'     => $reflector->getDocComment(),
            'attributes'     => self::_getAttributesInfo($reflector->getAttributes()),
            'inNamespace'    => $reflector->inNamespace(),
            'isAbstract'     => $reflector->isAbstract(),
            'isAnonymous'    => $reflector->isAnonymous(),
            //'isEnum'         => $reflector->isEnum(),
            'isFinal'        => $reflector->isFinal(),
            'isInstantiable' => $reflector->isInstantiable(),
            'isInterface'    => $reflector->isInterface(),
            'isInternal'     => $reflector->isInternal(),
            'isIterable'     => $reflector->isIterable(),
            'isTrait'        => $reflector->isTrait(),
            'isUserDefined'  => $reflector->isUserDefined(),
            'traits'         => array_combine($reflector->getTraitNames(), $reflector->getTraitNames()),
            'interfaces'     => array_combine($reflector->getInterfaceNames(), $reflector->getInterfaceNames()),
            'parent'         => $reflector->getParentClass() === false ? false : $reflector->getParentClass()->getname(),
            'methods'        => [],
        ];
        // Методы с константами
        try {
            $ret = array_merge($ret, [
                'isCloneable'       => $reflector->isCloneable(),
                'defaultProperties' => $reflector->getDefaultProperties(),
                'constants'         => $reflector->getConstants(),
            ]);
        } catch (\Error $e) {
            $ret = array_merge($ret, [
                'isCloneable'       => [],
                'defaultProperties' => [],
                'constants'         => [],
            ]);
        }

        // Информация о методах
        foreach ($reflector->getMethods() as $reflectorMethod) {
            $name = $reflectorMethod->getName();
            $mi   = [
                'name'          => $name,
                'docComment'    => $reflectorMethod->getDocComment(),
                'attributes'    => self::_getAttributesInfo($reflectorMethod->getAttributes()),
                'isAbstrac'     => $reflectorMethod->isAbstract(),
                'isConstructor' => $reflectorMethod->isConstructor(),
                'isDestructor'  => $reflectorMethod->isDestructor(),
                'isFinal'       => $reflectorMethod->isFinal(),
                'isPrivate'     => $reflectorMethod->isPrivate(),
                'isProtected'   => $reflectorMethod->isProtected(),
                'isPublic'      => $reflectorMethod->isPublic(),
                'isStatic'      => $reflectorMethod->isStatic(),
                'isClosure'     => $reflectorMethod->isClosure(),
                'isDeprecated'  => $reflectorMethod->isDeprecated(),
                'isGenerator'   => $reflectorMethod->isGenerator(),
                'isInternal'    => $reflectorMethod->isInternal(),
                'isUserDefined' => $reflectorMethod->isUserDefined(),
                'isVariadic'    => $reflectorMethod->isVariadic(),
                'args'          => [],
                'ret'           => self::_getTypeInfo($reflectorMethod->getReturnType()),
            ];
            // Аргументы
            $args = [];
            foreach ($reflectorMethod->getParameters() as $reflectorParameter) {
                $pi = [
                    'name'                    => $reflectorParameter->getName(),
                    'attributes'              => self::_getAttributesInfo($reflectorParameter->getAttributes()),
                    'allowsNull'              => $reflectorParameter->allowsNull(),
                    //'isArray'                 => $reflectorParameter->isArray(),
                    //'isCallable'              => $reflectorParameter->isCallable(),
                    'isDefaultValueAvailable' => $reflectorParameter->isDefaultValueAvailable(),
                    //'isDefaultValueConstant'  => $reflectorParameter->isDefaultValueConstant(),
                    'isOptional'              => $reflectorParameter->isOptional(),
                    'isPassedByReference'     => $reflectorParameter->isPassedByReference(),
                    'isVariadic'              => $reflectorParameter->isVariadic(),
                    'type'                    => self::_getTypeInfo($reflectorParameter->getType()),
                ];
                $args[] = $pi;
            }
            $mi['args'] = $args;
            // Имя метожа
            $ret['methods'][$name] = $mi;
        }
        // Вернуть информацию о классе
        return $ret;
    }
    //*/
    // Это файл PHP
    public static function isPhpFile(\SplFileInfo|string $spi): bool
    {
        // Если имя файла задано строкой
        if (is_string($spi)) {
            // то преобразовать его в объект
            $spi = new \SplFileInfo($spi);
        }
        return ($spi->isFile() && $spi->getExtension() == 'php');
    }
}
